<?php

namespace ReneSeindal;

if ( ! defined( 'ABSPATH' ) ) exit; // Exit if accessed directly

/************************************************************************
 *
 *	Schema.org data
 *
 ************************************************************************/


trait SEO_Schema {

    function schema_org( $object = NULL ) {
        if ( !$object )
            $object = get_queried_object();

        if ( empty( $object ) ) {
            if ( is_search() ) {
                $id = site_url( '?' . build_query( [ 's' => get_query_var( 's' ) ] ) );
                $schema = apply_filters( 'rs_seo_schema_search', [], $id );
            } else if ( is_404() )
                $schema = apply_filters( 'rs_seo_schema_404', [] );
            else
                $schema = apply_filters( 'rs_seo_schema_unknown', [] );
        }
        else if ( is_a( $object, 'WP_Post' ) ) {
            $id = get_permalink( $object ) . '#article';
            $schema = apply_filters( 'rs_seo_schema_article', [], $id, $object );
        }
        else if ( is_a( $object, 'WP_Post_Type' ) ) {
            $id = get_post_type_archive_link( $object->name );
            $schema = apply_filters( 'rs_seo_schema_collectionpage', [], $id,  $object );
        }
        else if ( is_a( $object, 'WP_Term' ) ) {
            $id = get_term_link( $object );
            $schema = apply_filters( 'rs_seo_schema_collectionpage', [], $id, $object );
        }
        else if ( is_a( $object, 'WP_User' ) ) {
            $id = get_author_posts_url( $object->ID);
            $schema = apply_filters( 'rs_seo_schema_profilepage', [], $id, $object );
        }

        return [
            '@context' => 'https://schema.org',
            '@graph' => array_values( $schema ),
        ];
    }

    // Merge a base schema with data from the settings
    function user_schema_merge( $base, $extras, $parms = [] ) {
        $name = $base['@type'];
        if ( is_array( $name ) )
            $name = $name[0];

        $schema = $this->get_user_schema( $name );

        if ( is_array( $schema ) ) {
            $subst = [
                '((SITE_NAME))' => get_bloginfo( 'name ' ),
                '((SITE_DESC))' => get_bloginfo( 'description ' ),
                '((SITE_URL))' => get_bloginfo( 'url ' ),
                '((SITE_LANG))' => get_bloginfo( 'language ' ),
            ];
            foreach ( $parms as $type => $parm ) {
                if ( empty( $parm ) ) continue;

                switch ( $type ) {
                case 'post':
                    $subst['((POST_ID))'] = $parm->ID;
                    $subst['((POST_NAME))'] = $parm->post_name;
                    $subst['((POST_TITLE))'] = $parm->post_title;
                    $subst['((POST_EXCERPT))'] = get_the_excerpt( $post );
                    $subst['((POST_TYPE))'] = $parm->post_type;
                    $subst['((POST_URL))'] = get_permalink( $parm );
                    break;
                case 'user':
                case 'author':
                    $subst['((USER_ID))'] = $parm->ID;
                    $subst['((USER_LOGIN))'] = $parm->user_login;
                    $subst['((USER_EMAIL))'] = $parm->user_email;
                    $subst['((USER_NAME))'] = $parm->display_name;
                    $subst['((USER_FIRST))'] = get_user_meta( $parm->ID, 'first_name', true );
                    $subst['((USER_LAST))'] = get_user_meta( $parm->ID, 'last_name', true );
                    $subst['((USER_DESC))'] = get_user_meta( $parm->ID, 'description', true );
                    $subst['((USER_URL))'] = get_author_posts_url( $parm->ID );
                    break;
                case 'post_type':
                    $subst['((COLL_NAME))'] = $parm->name;
                    $subst['((COLL_LABEL))'] = $parm->label;
                    $subst['((COLL_DESC))'] = $parm->description;
                    $subst['((COLL_URL))'] = get_permalink( $parm );
                    break;
                case 'term':
                    $subst['((COLL_NAME))'] = $parm->slug;
                    $subst['((COLL_LABEL))'] = $parm->name;
                    $subst['((COLL_DESC))'] = $parm->description;
                    $subst['((COLL_URL))'] = get_term_link( $parm );
                    break;
                case 'search':
                    $subst['((SEARCH))'] = $parm['search'];
                    break;
                }
            }

            foreach ( $subst as &$v )
                if ( str_contains( $v, "\n" ) )
                    $v = strtr( $v, "\n", ' ' );

            foreach ( $schema as $key => $value ) {
                if ( str_starts_with( $key, '@' ) )
                    continue;
                if ( empty( $base[$key] ) )
                    $base[$key] = $this->user_schema_merge_substitute( $value, $subst );
            }
        }

        // Additional fields which the user can override.
        foreach ( $extras as $key => $value )
            if ( empty( $base[$key] ) )
                $base[$key] = $value;


        return $base;
    }

    protected function user_schema_merge_substitute( $value, $subst ) {
        if ( is_string( $value ) )
            return str_replace( array_keys( $subst ), array_values( $subst ), $value );

        if ( is_array( $value ) ) {
            $new = [];
            foreach ( $value as $k => $v )
                $new[$k] = $this->user_schema_merge_substitute( $v, $subst );
            return $new;
        }

        return $value;
    }

    /************************************************************************
     *
     *	Article -- posts are articles
     *
     ************************************************************************/

    function do_rs_seo_schema_article_filter( $schema, $id, $post ) {
        if ( array_key_exists( $id, $schema ) )
            return $schema;

        $permalink = get_permalink( $post );
        $author = get_userdata( $post->post_author );

        $object = $this->user_schema_merge(
            [
                '@type' => 'Article',
                '@id' => $id,
                'isPartOf' => [
                    '@id' => $permalink,
                ],
                'author' => [
                    'name' => $author->display_name,
                    '@id' => site_url('#/schema/person/' . $author->user_login),
                ],
                'datePublished' => ( get_post_datetime( $post, 'date' ) )->format( 'c' ),
                'dateModified' => ( get_post_datetime( $post, 'modified' ) )->format( 'c' ),
                'mainEntityOfPage' => [
                    '@id' => $permalink,
                ],
                'wordCount' => str_word_count( strip_tags( $post->post_content ) ),
                'commentCount' => ( get_comment_count( $post->ID ) )['approved'],
                'publisher' => [
                    '@id' => site_url( '#organization' ),
                ],
                'image' => [
                    '@id' => $permalink . '#primaryimage',
                ],
            ],
            [
                'headline' => $post->post_title,
                'thumbnailUrl' => wp_get_attachment_url( get_post_thumbnail_id( $post ) ),
                'inLanguage' => get_bloginfo( 'language' ),
            ],
            [ 'post' => $post, 'user' => $author ]
        );

        $tags = get_the_tags( $post );
        if ( $tags && is_array( $tags ) )
            $object['keywords'] = array_map( fn($t) => $t->name, $tags );

        $cats = get_the_category( $post->ID );
        if ( $cats && is_array( $cats ) )
            $object['articleSection'] = array_map( fn($t) => $t->name, $cats );

        if ( 'open' === $post->comment_status ) {
            $object['potentialAction'] = [
                [
                    '@type' => 'CommentAction',
                    'name' => 'Comment',
                    'target' => [ $permalink . '#respond' ],
                ],
            ];
        }

        $schema[ $id ] = $object;

        $schema = apply_filters( 'rs_seo_schema_webpage', $schema, $object['isPartOf']['@id'], $post );
        $schema = apply_filters( 'rs_seo_schema_person', $schema, $object['author']['@id'], $author );
        $schema = apply_filters( 'rs_seo_schema_organization', $schema, $object['publisher']['@id'] );
        $schema = apply_filters( 'rs_seo_schema_imageobject', $schema, $object['image']['@id'], get_post_thumbnail_id( $post ) );

        return $schema;
    }

    /************************************************************************
     *
     *	CollectionPage -- archives for taxonomy terms and post types
     *
     ************************************************************************/

    function do_rs_seo_schema_collectionpage_filter( $schema, $id, $term_or_type ) {
        if ( array_key_exists( $id, $schema ) )
            return $schema;

        switch ( get_class( $term_or_type ) ) {
        case 'WP_Term':
            $merge = [ 'term' => $term_or_type ];
            $term = $term_or_type;
            $name = $term->name;
            $url = get_term_link( $term );

            $taxonomy = get_taxonomy( $term->taxonomy );
            $query = [
                'post_type' => $taxonomy->object_type,
                'tax_query' => [
                    [
                        'taxonomy' => $term->taxonomy,
                        'terms' => $term->term_id,
                        'field' => 'term_id',
                    ]
                ],
                'numberposts' => 1,
            ];
            break;

        case 'WP_Post_Type':
            $merge = [ 'post_type' => $term_or_type ];
            $post_type = $term_or_type;
            $name = $post_type->labels->name;
            $url = get_post_type_archive_link( $post_type->name );
            $query = [
                'post_type' => $post_type->name,
                'numberposts' => 1,
            ];
            break;

        default:
            return $schema;
        }

        $page = get_query_var( 'paged', 1 );
        if ( $page > 1 )
            $url = get_pagenum_link( $page );



        $object = $this->user_schema_merge(
            [
                '@type' => 'CollectionPage',
                '@id' => $id,
                'url' => $url,
                'isPartOf' => [
                    '@id' => site_url('#website'),
                ],
                'breadcrumb' => [
                    '@id' => $url . '#breadcrumb',
                ],
            ],
            [
                'name' => $name,
                'inLanguage' => get_bloginfo( 'language' ),
            ],
            $merge
        );

        $posts = get_posts( $query );
        if ( $posts ) {
            $object['thumbnailUrl'] = wp_get_attachment_url( get_post_thumbnail_id( $posts[0] ) );
            $object['primaryImageOfPage'] = $object['image'] = [ '@id' => $url . '#primaryimage' ];
        }

        $schema[ $id ] = $object;

        $schema = apply_filters( 'rs_seo_schema_website', $schema, $object['isPartOf']['@id'] );
        if ( !empty( $object['primaryImageOfPage']['@id'] ) )
            $schema = apply_filters( 'rs_seo_schema_imageobject', $schema, $object['primaryImageOfPage']['@id'], $posts[0] );
        if ( !empty( $object['image']['@id'] ) )
            $schema = apply_filters( 'rs_seo_schema_imageobject', $schema, $object['image']['@id'], $posts[0] );

        $schema = apply_filters( 'rs_seo_schema_breadcrumblist', $schema, $object['breadcrumb']['@id'], $term_or_type );

        return $schema ;
    }

    /************************************************************************
     *
     *	ProfilePage -- author archives
     *
     ************************************************************************/

    function do_rs_seo_schema_profilepage_filter( $schema, $id, $user ) {
        if ( array_key_exists( $id, $schema ) )
            return $schema;

        $url = $id;

        $object = $this->user_schema_merge(
            [
                '@type' => 'ProfilePage',
                '@id' => $id,
                'url' => $url,
                'isPartOf' => [
                    '@id' => site_url('#website'),
                ],
                'author' => [
                    'name' => $user->display_name,
                    '@id' => site_url('#/schema/person/' . $user->user_login),
                ],
                'breadcrumb' => [
                    '@id' => $url . '#breadcrumb',
                ],
                'potentialAction' => [
                    [
                        '@type' => 'ReadAction',
                        'target' => [ $url ],
                    ],
                ]
            ],
            [
                'name' => $user->display_name,
                'inLanguage' => get_bloginfo( 'language' ),
            ],
            [ 'user' => $user ]
        );

        $posts = get_posts( [ 'author' => $user->ID, 'post_type' => 'any', 'numberposts' => 1, ] );
        if ( $posts ) {
            $object['thumbnailUrl'] = wp_get_attachment_url( get_post_thumbnail_id( $posts[0] ) );
            $object['primaryImageOfPage'] = $object['image'] = [ '@id' => $url . '#primaryimage' ];
        }

        $schema[ $id ] = $object;

        $schema = apply_filters( 'rs_seo_schema_website', $schema, $object['isPartOf']['@id'] );
        if ( !empty( $object['primaryImageOfPage']['@id'] ) )
            $schema = apply_filters( 'rs_seo_schema_imageobject', $schema, $object['primaryImageOfPage']['@id'], $posts[0] );
        if ( !empty( $object['image']['@id'] ) )
            $schema = apply_filters( 'rs_seo_schema_imageobject', $schema, $object['image']['@id'], $posts[0] );

        $schema = apply_filters( 'rs_seo_schema_person', $schema, $object['author']['@id'], $user );
        $schema = apply_filters( 'rs_seo_schema_breadcrumblist', $schema, $object['breadcrumb']['@id'], $user );

        return $schema ;
    }

    /************************************************************************
     *
     *	SearchResultsPage -- search results
     *
     ************************************************************************/

    function do_rs_seo_schema_search_filter( $schema, $id ) {
        if ( array_key_exists( $id, $schema ) )
            return $schema;

        $url = $id;

        $object = $this->user_schema_merge(
            [
                '@type' => [ 'SearchResultsPage', 'CollectionPage' ],
                '@id' => $id,
                'url' => $url,
                'isPartOf' => [
                    '@id' => site_url('#website'),
                ],
                'breadcrumb' => [
                    '@id' => $url . '#breadcrumb',
                ],
            ],
            [
                'name' => get_query_var( 's' ),
                'inLanguage' => get_bloginfo( 'language' ),
            ],
            [ 'search' => [ 'search' => get_query_var( 's' ) ] ]
        );

        $schema[ $id ] = $object;

        $schema = apply_filters( 'rs_seo_schema_website', $schema, $object['isPartOf']['@id'] );
        $schema = apply_filters( 'rs_seo_schema_breadcrumblist', $schema, $object['breadcrumb']['@id'], NULL );

        return $schema ;
    }

    /************************************************************************
     *
     *	WebPages - contain Article, CollectionPage and SearchResultsPage
     *
     ************************************************************************/

    function do_rs_seo_schema_webpage_filter( $schema, $id, $post ) {
        if ( array_key_exists( $id, $schema ) )
            return $schema;

        $permalink = get_permalink( $post );

        $object = $this->user_schema_merge(
            [
                '@type' => 'WebPage',
                '@id' => $id,
                'url' => $permalink,
                'isPartOf' => [
                    '@id' => site_url( '#website' ),
                ],
                'primaryImageOfPage' => [
                    '@id' => site_url( '#primaryimage' ),
                ],
                'image' => [
                    '@id' => site_url( '#primaryimage' ),
                ],
                'thumbnailUrl' => wp_get_attachment_url( get_post_thumbnail_id( $post ) ),
                'datePublished' => ( get_post_datetime( $post, 'date' ) )->format( 'c' ),
                'dateModified' => ( get_post_datetime( $post, 'modified' ) )->format( 'c' ),
                'breadcrumb' => [
                    '@id' => site_url( '#breadcrumb' ),
                ],
                'potentialAction' => [
                    [
                        '@type' => 'ReadAction',
                        'target' => [ $permalink ],
                    ],
                ],
            ],
            [
                'inLanguage' => get_bloginfo( 'language' ),
                'name' => $post->post_title,
                'description' => get_the_excerpt( $post ),
            ],
            [ 'post' => $post ]
        );

        $schema[ $id ] = $object;

        $schema = apply_filters( 'rs_seo_schema_website',			$schema, $object['isPartOf']['@id'] );
        $schema = apply_filters( 'rs_seo_schema_imageobject',		$schema, $object['image']['@id'], $post );
        $schema = apply_filters( 'rs_seo_schema_breadcrumblist',	$schema, $object['breadcrumb']['@id'], $post );

        return $schema;
    }

    /************************************************************************
     *
     *	WebSite -- contains WebPages
     *
     ************************************************************************/

    function do_rs_seo_schema_website_filter( $schema, $id ) {
        if ( array_key_exists( $id, $schema ) )
            return $schema;

        $object = $this->user_schema_merge(
            [
                '@type' => 'WebSite',
                '@id' => $id,
                'url' => site_url(),
                'publisher' => [
                    '@id' => site_url( '#organization' ),
                ],
                'potentialAction' => [
                    [
                        '@type' => 'SearchAction',
                        'target' => [
                            '@type' => 'EntryPoint',
                            'urlTemplate' => site_url( '?s={search_term_string}' ),
                        ],
                        'query-input' => 'required name=search_term_string',
                    ],
                ],
            ],
            [
                'name' => get_bloginfo( 'name' ),
                'description' => get_bloginfo( 'description' ),
                'inLanguage' => get_bloginfo( 'language' ),
            ]
        );

        $schema[ $object['@id'] ] = $object;

        $schema = apply_filters( 'rs_seo_schema_organisation', $schema, $object['publisher']['@id'] );

        return $schema;
    }


    /************************************************************************
     *
     *	Persons -- authors of Articles and ProfilePages
     *
     ************************************************************************/

    function do_rs_seo_schema_person_filter( $schema, $id, $user ) {
        if ( array_key_exists( $id, $schema ) )
            return $schema;

        $object = $this->user_schema_merge(
            [
                '@type' => 'Person',
                '@id' => $id,
                'name' => $user->display_name,
                'url' => get_author_posts_url( $user->ID ),
            ],
            [
                'email' => $user->user_email,
                'description' => $user->description,
            ],
            [ 'user' => $user ]
        );

        $address = strtolower( trim( $user->user_email ) );
        $hash = hash( 'sha256', $address );
        $gravatar = 'https://www.gravatar.com/avatar/' . $hash;

        $object['image'] = [
            '@type' => 'ImageObject',
            'inLanguage' => get_bloginfo( 'language' ),
            '@id' => site_url( '#/schema/person/image/' ),
            'url' => $gravatar,
            'contentUrl' => $gravatar,
            'caption' => $user->display_name,
        ];

        $user_social = get_user_meta( $user->ID, 'rs_seo_basics_sameAs', true );
        if ( !empty( $user_social ) ) {
            $sameAs = [];
            foreach ( explode( "\n", $user_social ) as $link ) {
                $link = trim( $link );
                if ( $link && wp_http_validate_url( $link ) )
                    $sameAs[] = $link;
            }

            if ( $sameAs )
                $object['sameAs'] = $sameAs;
        }

        $schema[ $id ] = $object;

        return $schema;
    }

    /************************************************************************
     *
     *	Organizations have WebSites
     *
     ************************************************************************/

    function do_rs_seo_schema_organisation_filter( $schema, $id ) {
        if ( array_key_exists( $id, $schema ) )
            return $schema;

        $object = $this->user_schema_merge(
            [
                '@type' => 'Organization',
                '@id' => $id,
                'url' => site_url(),
            ],
            [
                'name' => get_bloginfo( 'name' ),
            ]
        );


        if ( empty( $object['logo'] ) ) {
            $site_icon_id = (int) get_option( 'site_icon' );
            if ( $site_icon_id ) {
                $site_icon_url = wp_get_attachment_image_url( $site_icon_id, 'full' );

                $icon_id = site_url( '#/schema/logo/image/' );
                $logo = [
                    '@type' => 'ImageObject',
                    '@id' => $icon_id,
                    'inLanguage' => get_bloginfo( 'language' ),
                    'url' => $site_icon_url,
                    'contentUrl' => $site_icon_url,
                    'caption' => get_bloginfo( 'name' ),
                ];

                $meta = wp_get_attachment_metadata( $site_icon_id );
                if ( is_array( $meta ) ) {
                    $logo['width'] = $meta['width'];
                    $logo['height'] = $meta['height'];
                }

                $object['logo'] = $logo;
                $object['image'] = [ '@id' => $icon_id ];
            }
        }

        $schema[ $id ] = $object;

        return $schema;
    }

    /************************************************************************
     *
     *	ImageObjects for featured images
     *
     ************************************************************************/

    function do_rs_seo_schema_imageobject_filter( $schema, $id, $image_id ) {
        if ( array_key_exists( $id, $schema ) )
            return $schema;

        if ( is_a( $image_id, 'WP_Post' ) )
            $image_id = get_post_thumbnail_id( $image_id );

        $object = [
            '@type' => 'ImageObject',
            '@id' => $id,
            'inLanguage' => get_bloginfo( 'language' ),
        ];

        $url = wp_get_attachment_url( $image_id );
        if ( $url ) {
            $object['url'] = $url;
            $object['contentUrl'] = $url;
        }

        $meta = wp_get_attachment_metadata( $image_id );
        if ( $meta ) {
            $object['width'] = $meta['width'];
            $object['height'] = $meta['height'];
        }

        $alt = get_post_meta( $image_id, '_wp_attachment_image_alt', true );
        if ( $alt )
            $object['caption'] = $alt;

        $schema[ $id ] = $object;

        return $schema;
    }


    /************************************************************************
     *
     *	Most pages have BreadCrumbs
     *
     ************************************************************************/

    function do_rs_seo_schema_breadcrumblist_filter( $schema, $id, $object ) {
        if ( array_key_exists( $id, $schema ) )
            return $schema;

        // Find name of this object

        $name = NULL;
        if ( is_object( $object ) ) {
            switch ( get_class( $object ) ) {
            case 'WP_Post':
                $name = $object->post_title;
                break;
            case 'WP_Post_Type':
                $name = $object->labels->name;
                break;
            case 'WP_Term':
                $name = $object->name;
                break;
            case 'WP_User':
                $name = 'Author: ' . $object->display_name;
                break;
            }
        } else {
            if ( is_search() )
                $name = sprintf( 'You searched for %s', get_query_var( 's' ) );
        }

        if  ( empty( $name ) )
            $name = __( 'Unknown object', 'rs-seo-basics' );


        // Build reserve path of 'name' and 'item' elements only

        $path = [
            [ 'name' => $name, ]
        ];

        if ( is_a( $object, 'WP_Post' ) ) {
            $parent = $object;
            while ( $parent->post_parent > 0 ) {
                $parent = get_post( $parent->post_parent );
                $path[] = [
                    'name' => $parent->post_title,
                    'item' => get_permalink( $parent ),
                ];
            }

            if ( 'post' === $object->post_type ) {
                $path[] = [
                    'name' => 'Blog',
                    'item' => get_post_type_archive_link( $object->post_type ),
                ];
            } else {
                $post_type = get_post_type_object( $object->post_type );
                if ( $post_type->has_archive ) {
                    $path[] = [
                        'name' =>  $post_type->labels->name,
                        'item' => get_post_type_archive_link( $post_type->name ),
                    ];
                }
            }
        }

        $path[] = [
            'name' => 'Home',
            'item' => site_url(),
        ];

        // Reverse list and add the '@type' and 'position' elements

        $path = array_reverse( $path );
        foreach ( $path as $k => &$v )
            $v = array_merge( [ '@type' => 'ListItem', 'position' => $k+1 ], $v );

        // Wrap it up

        $object =    [
            '@type' => 'BreadcrumbList',
            '@id' => $id,
            'itemListElement' => $path,
        ];

        $schema[ $id ] = $object;

        return $schema;
    }
}
